@@ -3,7 +3,10 @@ import json
 import re
 from collections import namedtuple
 
-import requests
+from requests import request
+from requests.exceptions import HTTPError
+from urllib3 import disable_warnings
+from urllib3.exceptions import InsecureRequestWarning
 import sys
 import time
 
@@ -31,6 +34,9 @@ class LogicHubAPI:
     __case_prefix = None
     __fields = None
     __notebooks = None
+    __hostname = None
+    __user_id = None
+    __user_role = None
 
     # Variables used only for API token auth
     __api_key = None
@@ -43,12 +49,16 @@ class LogicHubAPI:
     __session_cookie = None
 
     def __init__(self, hostname, api_key=None, username=None, password=None, verify_ssl=True, cache_seconds=None, default_timeout=None, **kwargs):
-        # First store key variables and then determine whether they are valid
-        self.__api_key = api_key.strip() if api_key else None
-        self.__username = username.strip() if username else None
-        self.__password = password.strip() if password else None
-        assert api_key or (username and password), "Must provide either an API key or a username and password"
-        assert not (api_key and password), "Using both an API token and a password is not supported"
+        if not hostname:
+            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="No hostname provided")
+        if not api_key and not (username and password):
+            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="Must provide either an API key or a username and password")
+        if api_key and password:
+            raise exceptions.validation.InputValidationError(input_var=None, action_description="instantiation", message="Using both an API token and a password is not supported")
+        self.__hostname = hostname
+        self.__api_key = api_key or None
+        self.__username = username or None
+        self.__password = password or None
 
         self.cache_seconds = int(cache_seconds or LogicHubAPI.DEFAULT_CACHE_SECONDS)
         if self.cache_seconds is None or self.cache_seconds < 0:
@@ -66,8 +76,6 @@ class LogicHubAPI:
         if not self.verify_ssl:
             # Disable certificate warnings
             self.log.debug("Disabling SSL warnings...")
-            from urllib3 import disable_warnings
-            from urllib3.exceptions import InsecureRequestWarning
             disable_warnings(InsecureRequestWarning)
 
         if self.__api_key:
@@ -76,8 +84,9 @@ class LogicHubAPI:
             self.auth_type = 'password'
             self.__credentials = {"email": self.__username, "password": self.__password}
 
-        self.url = URLs(hostname)
-        self.url._version = self.version
+        __init_version = self.kwargs.pop('init_version', None)
+        self.url = URLs(hostname, init_version=__init_version)
+        self.__set_version(__init_version)
         _ = atexit.register(self.close)
 
     def __enter__(self):
@@ -118,7 +127,6 @@ class LogicHubAPI:
     def fields(self):
         if not self.__get_cached_object(self.__fields):
             _ = self.list_fields()
-            self.log.debug(f"Updated cached property: fields (timeout: {self.cache_seconds} sec")
         return self.__fields.value
 
     @fields.setter
@@ -130,18 +138,30 @@ class LogicHubAPI:
         for f in self.fields:
             if f.get('fieldName') == 'lh_linked_alerts':
                 return f
-        raise exceptions.Validation.VersionMinimumNotMet(min_version='m86', feature_label='linked alerts')
+        raise exceptions.validation.VersionMinimumNotMet(min_version='m86', feature_label='linked alerts')
 
     @property
     def version(self):
         if not self.__get_cached_object(self.__version):
             _ = self.get_version_info()
-            self.log.debug(f"LogicHub version: m{self.__version}")
         return self.__version.value
 
-    @version.setter
-    def version(self, value):
-        self.__version = cached_obj(int(time.time()), value)
+    @property
+    def major_version(self):
+        return int(float(self.version))
+
+    @property
+    def minor_version(self):
+        return int(re.sub(r'^.*?(\d+)$', '$1', self.version))
+
+    def __set_version(self, value):
+        if value is None:
+            _ = self.get_version_info()
+            return
+        self.__version = cached_obj(int(time.time()), helpers.format_version(value))
+        self.log.debug(f"LogicHub version: {self.version}")
+        # Update the version attribute in self.url for use in other calls since the URLs class does not make any such calls on its own
+        self.url._current_version = self.version
 
     @property
     def case_prefix(self):
@@ -149,6 +169,28 @@ class LogicHubAPI:
             _ = self.cases_get_prefix()
         return self.__case_prefix.value
 
+    @property
+    def session_hostname(self):
+        return self.__hostname
+
+    @property
+    def session_username(self):
+        if not self.__username:
+            _ = self.me()
+        return self.__username
+
+    @property
+    def session_user_id(self):
+        if not self.__user_id:
+            _ = self.me()
+        return self.__user_id
+
+    @property
+    def session_user_role(self):
+        if not self.__user_role:
+            _ = self.me()
+        return self.__user_role
+
     @case_prefix.setter
     def case_prefix(self, value):
         self.__case_prefix = cached_obj(int(time.time()), value)
@@ -170,9 +212,9 @@ class LogicHubAPI:
 
     @property
     def session_cookie(self):
-        # ToDo Replace this later, but for now include this until I'm sure the update is all ready
+        # This should only ever come up w/ password auth, so if this has been invoked when there is an API key then throw an exception
         if self.__api_key:
-            raise Exception('Session cookie should not be invoked')
+            raise exceptions.LhBaseException('Session cookie should not be invoked')
         if not self.__session_cookie:
             self.login()
         return self.__session_cookie
@@ -191,29 +233,29 @@ class LogicHubAPI:
         # Group all tests for status code 401
         if response_obj.status_code == 401:
             if self.auth_type == 'password':
-                raise exceptions.Auth.PasswordAuthFailure
+                raise exceptions.auth.PasswordAuthFailure
             else:
                 if 'Unauthorized for URL' in response_obj.text or 'token is not allowed with this endpoint' in response_obj.text:
-                    raise exceptions.Auth.APIAuthNotAuthorized(url)
+                    raise exceptions.auth.APIAuthNotAuthorized(url)
                 else:
-                    raise exceptions.Auth.APIAuthFailure(url)
+                    raise exceptions.auth.APIAuthFailure(url)
 
         # Group all tests for status code 400
         elif response_obj.status_code == 400:
             if self.auth_type == 'password':
                 if url == self.url.login and 'Username/Password not valid' in self.last_response_text:
-                    raise exceptions.Auth.PasswordAuthFailure
+                    raise exceptions.auth.PasswordAuthFailure
             if 'batch' in url and re.search(r'controllers.BatchController.legacyPost.*?For input string', response_obj.text):
                 _id = re.search(r'POST /demo/batch-(\d+)', response_obj.text)
                 _id = _id.groups()[0] if _id else None
-                raise exceptions.BatchNotFound(_id)
+                raise exceptions.app.BatchNotFound(_id)
 
         # Group all tests for status code 500
         elif response_obj.status_code == 500:
             if 'Unable to find batch with id' in response_obj.text:
                 _id = re.search(r'Unable to find batch with id batch-(\d+)', response_obj.text)
                 _id = _id.groups()[0] if _id else None
-                raise exceptions.BatchNotFound(_id)
+                raise exceptions.app.BatchNotFound(_id)
 
         response_obj.raise_for_status()
 
@@ -245,7 +287,7 @@ class LogicHubAPI:
             json_body = {}
 
         kwargs = {k: v for k, v in {**{"params": params, "data": body, "json": json_body, "headers": headers}, **(kwargs or {})}.items() if v}
-        response = requests.request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)
+        response = request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)
 
         # Only use reauth/automatic login if password auth is used
         if self.auth_type == 'password':
@@ -253,7 +295,7 @@ class LogicHubAPI:
             if response.status_code == 401 and reauth and "not logged in" in response.text:
                 self.log.debug(f"Session expired; logging in and retrying")
                 self.login(force_new_session=True)
-                response = requests.request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)
+                response = request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)
 
         # Store last response before testing whether the call was successful
         self.last_response_status = response.status_code
@@ -278,20 +320,21 @@ class LogicHubAPI:
         self.log.debug("Login successful")
 
     def logout(self):
+        self.log.debug("Logout requested")
         if self.auth_type != 'password':
             raise exceptions.LhBaseException('Login process is only used for password auth')
         if not self.session_cookie:
             self.log.debug("Already logged out; logout skipped")
             return
 
-        self.log.debug("Logout requested")
+        self.log.debug("Issuing logout command")
         _response = self._http_request(url=self.url.logout, method="POST", test_response=False, reauth=False, timeout=self._http_timeout_login)
         try:
             _response.raise_for_status()
-        except requests.exceptions.HTTPError as err:
-            self.log.warn(f"Logout failed with HTTP error: {str(err)}")
+        except HTTPError as err:
+            self.log.warning(f"Logout failed with HTTP error: {str(err)}")
         except Exception as err:
-            self.log.warn(f"Logout failed with UNKNOWN error: {repr(err)}")
+            self.log.warning(f"Logout failed with UNKNOWN error: {repr(err)}")
         else:
             self.log.debug(f"Logout successful")
             self.session_cookie = None
@@ -337,6 +380,7 @@ class LogicHubAPI:
         limit = int(limit or 1000)
         offset = int(offset or 0)
         params = {"fields": "*", "pageSize": limit, "after": offset, "cachedOnly": "true"}
+        self.log.debug("Fetching batch results")
         response = self._http_request(url=self.url.batch_results_by_id.format(int(batch_id)), params=params)
         return response.json()
 
@@ -356,6 +400,7 @@ class LogicHubAPI:
             "after": int(offset or 0),
         }
         body = {"status": statuses or [], "excludeBatchesWithZeroEvents": exclude_empty_results}
+        self.log.debug("Fetching batches")
         response = self._http_request(method="POST", url=self.url.stream_batches.format(stream_id), params=params, body=body)
         return response.json()
 
@@ -375,7 +420,7 @@ class LogicHubAPI:
                 connection_list.append(f"connection-{connection}")
 
         headers = {"Content-Type": "application/json"}
-        # response = self._http_request(method="POST", url=self.url.connection_status, headers=headers, body=json.dumps(connection_list))
+        self.log.debug("Fetching connection status")
         response = self._http_request(method="POST", url=self.url.connection_status, headers=headers, body=connection_list)
         return response.json()
 
@@ -405,6 +450,7 @@ class LogicHubAPI:
             url = self.url.custom_list_data
             method = "GET"
             payload = None
+        self.log.debug("Fetching custom list data")
         response = self._http_request(
             method=method,
             url=url,
@@ -416,11 +462,12 @@ class LogicHubAPI:
         try:
             results = results["result"]
         except Exception:
-            raise exceptions.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
+            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
         return results
 
     # ToDo Token auth not supported as of 2022-03-09 (m92)
     def get_dashboard_data(self, dashboard_id):
+        self.log.debug("Fetching dashboard data")
         response = self._http_request(url=self.url.dashboard_data.format(dashboard_id))
         return response.json()
 
@@ -429,44 +476,38 @@ class LogicHubAPI:
         rule_sets = self.list_rule_sets()
         rule_set = [x for x in rule_sets if x['name'] == name]
         if not rule_set:
-            raise exceptions.RuleSetNotFound(f"No rule set found matching name: {name}")
+            raise exceptions.app.RuleSetNotFound(input_var=name)
         rule_set = rule_set[0]
         rule_set['rules'] = self.get_rules_for_rule_set(rule_set['id'])
         return rule_set
 
     # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
     def get_rules_for_rule_set(self, rule_set):
-        rule_set = helpers.sanitize_input_rule_set_id(rule_set)
+        rule_set = helpers.format_rule_set_id(rule_set)
         params = {
             "fields": "name,isPublic,rules[filter,values*1000,score]*1000",
         }
+        self.log.debug("Fetching rules")
         response = self._http_request(url=self.url.rule_set.format(rule_set), params=params)
         results = response.json()
         try:
             results = results["result"]["rules"]["data"]
         except Exception:
-            raise exceptions.UnexpectedOutput("API response does not match the expected schema for listing rules of a rule set")
+            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing rules of a rule set")
         return results
 
     def get_stream_by_id(self, stream_id: int):
         headers = {"Accept": "application/json"}
-        # ToDo Revisit this. A better approach might be:
-        # try:
-        #     response = self._http_request(url=self.url.stream_by_id.format(stream_id), headers=headers)
-        #     return response.json()
-        # except requests.exceptions.HTTPError:
-        #     if self.last_response_status == 400 and 'Cannot find entity for id StreamId' in self.last_response_text:
-        #         raise exceptions.StreamNotFound(stream_id)
+        self.log.debug("Fetching stream")
         response = self._http_request(url=self.url.stream_by_id.format(stream_id), headers=headers, test_response=False)
-        try:
+        if response.status_code == 400 and 'Cannot find entity for id StreamId' in response.text:
+            raise exceptions.app.StreamNotFound(stream_id)
+        else:
             response.raise_for_status()
-        except requests.exceptions.HTTPError:
-            if response.status_code == 400 and 'Cannot find entity for id StreamId' in response.text:
-                raise exceptions.StreamNotFound(stream_id)
         return response.json()
 
     def get_version_info(self):
-        self.log.debug("Fetching LogicHub version")
+        self.log.debug("Fetching LogicHub version info")
         try:
             response = self._http_request(
                 url=self.url.version,
@@ -478,20 +519,21 @@ class LogicHubAPI:
         else:
             # Update version information any time this api call is run successfully
             self.version_info = response_dict
-            self.version = float(re.match("m(.*)", self.version_info["version"]).group(1))
+            self.__set_version(self.version_info)
         return response_dict
 
     def get_workflow_by_id(self, workflow_id: int):
         assert isinstance(workflow_id, int), "Workflow ID must be an integer"
+        self.log.debug("Fetching workflow")
         response = self._http_request(method="GET", url=self.url.case_status_workflow_by_id.format(workflow_id))
         return response.json()
 
     # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
     # ToDo Create an action method for this, and update the beta integration action to use it
     def add_scoring_rule(self, rule_set, field_mappings, score: float or str):
-        rule_set = helpers.sanitize_input_rule_set_id(rule_set)
+        rule_set = helpers.format_rule_set_id(rule_set)
         field_mappings = helpers.sanitize_input_rule_field_mappings(field_mappings)
-        score = helpers.sanitize_input_rule_score(score)
+        score = helpers.format_rule_score(score)
         headers = {"Content-Type": "application/json"}
         body = {
             "method": "addRule",
@@ -500,15 +542,17 @@ class LogicHubAPI:
                 "score": score
             }
         }
+        self.log.debug("Issuing new scoring rule")
         response = self._http_request(url=self.url.rule_set.format(rule_set), method="POST", body=body, headers=headers)
         results = response.json()
         try:
             results = results["result"]
         except Exception:
-            raise exceptions.UnexpectedOutput("API response does not match the expected schema for adding a scoring rule")
+            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for adding a scoring rule")
         return results
 
     def export_playbook(self, playbook_id, test_response=True):
+        self.log.debug(f"Exporting playbook {playbook_id}")
         response = self._http_request(url=self.url.flow_export.format(playbook_id), test_response=test_response)
         return response.json()
 
@@ -520,6 +564,7 @@ class LogicHubAPI:
         :return:
         """
         limit = int(limit if limit and limit > 0 else 999999999)
+        self.log.debug("Executing LogicHub command")
         response = self._http_request(
             url=self.url.command_execute,
             method="POST",
@@ -565,7 +610,7 @@ class LogicHubAPI:
 
         try:
             response.raise_for_status()
-        except requests.exceptions.HTTPError:
+        except HTTPError:
             message = f"Request failed with status code {response.status_code}\n\nText:\n\n"
             try:
                 message += json.dumps(response.json(), indent=2)
@@ -574,25 +619,43 @@ class LogicHubAPI:
             self.log.fatal(message)
         return response.json()
 
-    def list_baselines(self, limit=None, offset=0):
+    def list_baselines(self, limit=None, offset=0, filters=None):
         """
         List all Baselines
 
         :param limit: None by default, although the LogicHub UI defaults to 25
         :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
+        :param filters: Optional: Advanced search filters (list of dicts)
         :return:
         """
         limit = int(limit if limit and limit > 0 else 999999999)
         params = {"libraryView": "all"}
-        body = {"filters": [], "offset": offset or 0, "pageSize": limit, "sortColumn": "lastUpdated", "sortOrder": "DESC"}
+        body = {"filters": filters or [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
+        self.log.debug("Fetching baselines")
         response = self._http_request(method="POST", url=self.url.baselines, params=params, body=body)
         return response.json()
 
+    def list_commands(self, limit=None, offset=0, filters=None):
+        """
+        List all Commands
+
+        :param limit: None by default, although the LogicHub UI defaults to 25
+        :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
+        :param filters: Optional: Advanced search filters (list of dicts)
+        :return:
+        """
+        limit = int(limit if limit and limit > 0 else 999999999)
+        params = {"libraryView": "all"}
+        body = {"filters": filters or [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
+        self.log.debug("Fetching commands")
+        response = self._http_request(method="POST", url=self.url.commands, params=params, body=body)
+        return response.json()
+
     def list_connections(self, filters=None, limit=None, offset=0):
         """
         List all connections
 
-        :param filters: Optional: Advanced search
+        :param filters: Optional: Advanced search filters (list of dicts)
         :param limit: None by default, although the LogicHub UI defaults to 25
         :param offset: Used for pagination if you want to pull in chunks, this sets the page number to pull
         :return:
@@ -601,6 +664,7 @@ class LogicHubAPI:
         params = {"libraryView": "all"}
         headers = {"Content-Type": "application/json"}
         body = {"filters": filters or [], "offset": offset, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
+        self.log.debug("Fetching connections")
         response = self._http_request(method="POST", url=self.url.connections, headers=headers, params=params, body=body)
         return response.json()
 
@@ -620,6 +684,7 @@ class LogicHubAPI:
         if search_text:
             filters.append({"searchText": search_text})
 
+        self.log.debug("Fetching custom lists")
         response = self._http_request(
             method="POST",
             url=self.url.custom_lists,
@@ -638,15 +703,17 @@ class LogicHubAPI:
             try:
                 _ = results["result"]["data"]
             except Exception:
-                raise exceptions.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
+                raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing custom lists")
         return results
 
     # ToDo Token auth not supported as of 2022-03-09 (m92)
     def list_dashboards(self):
+        self.log.debug("Fetching dashboards")
         response = self._http_request(url=self.url.dashboards)
         return response.json()
 
     def list_dashboards_with_widgets(self):
+        self.log.debug("Fetching dashboards with widgets")
         response = self._http_request(url=self.url.dashboards_and_widgets)
         return response.json()
 
@@ -659,6 +726,7 @@ class LogicHubAPI:
         # Not sure when this one stopped working, but it worked somewhere in at least m66 and didn't work any more in m70
         limit = int(limit if limit and limit > 0 else 999999999)
         params = {"limit": limit}
+        self.log.debug("Fetching event types")
         response = self._http_request(url=self.url.event_types, params=params)
         results = response.json()
         if results.get("result"):
@@ -678,6 +746,7 @@ class LogicHubAPI:
         params = {"libraryView": "all"}
         # Confirm the right term for "name" as the sort column and change this sort order
         body = {"filters": [], "pageSize": limit, "offset": offset or 0, "sortColumn": "lastUpdated", "sortOrder": "DESC"}
+        self.log.debug("Fetching event types")
         response = self._http_request(method="POST", url=self.url.event_types, params=params, body=body)
         results = response.json()
         if results.get("result"):
@@ -693,21 +762,25 @@ class LogicHubAPI:
         :return:
         """
         limit = int(limit if limit and limit > 0 else 999999999)
-        if self.version < 70:
+        # ToDo Remove this old version when it's safe to assume no one still uses m70
+        if self.major_version < 70:
             return self.__list_event_types_v1(limit=limit)
         return self.__list_event_types_v2(limit=limit)
 
     def list_fields(self, params: dict = None, **kwargs):
         params = params or {"systemFields": "true", "pageSize": 9999, "after": 0}
+        self.log.debug("Fetching fields")
         response = self._http_request(method="GET", url=self.url.fields, params=params, **kwargs)
         self.fields = response.json()
         return response.json()
 
     def list_integrations(self):
+        self.log.debug("Fetching integrations")
         response = self._http_request(method="GET", url=self.url.integrations)
         return response.json()
 
     def list_ml_models(self):
+        self.log.debug("Fetching ML models")
         response = self._http_request(
             method="GET",
             url=self.url.ml_models,
@@ -716,6 +789,7 @@ class LogicHubAPI:
         return response.json()
 
     def list_modules(self):
+        self.log.debug("Fetching modules")
         response = self._http_request(
             method="GET",
             url=self.url.modules,
@@ -742,6 +816,7 @@ class LogicHubAPI:
         }
         if search_string:
             body['filters'] = [{"searchText": search_string}]
+        self.log.debug("Fetching notebooks")
         response = self._http_request(
             url=self.url.notebooks_list,
             method="POST",
@@ -764,6 +839,7 @@ class LogicHubAPI:
         """
         limit = int(limit if limit and limit > 0 else 999999999)
         body = {"filters": [], "offset": offset or 0, "pageSize": limit, "sortColumn": "name", "sortOrder": "ASC"}
+        self.log.debug("Fetching playbooks")
         response = self._http_request(
             url=self.url.playbooks_list,
             method="POST",
@@ -784,12 +860,13 @@ class LogicHubAPI:
             "fields": "name,isPublic",
             "pageSize": limit,
         }
+        self.log.debug("Fetching rule sets")
         response = self._http_request(url=self.url.rule_sets, params=params)
         results = response.json()
         try:
             results = results["result"]["data"]
         except Exception:
-            raise exceptions.UnexpectedOutput("API response does not match the expected schema for listing rule sets")
+            raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing rule sets")
         return results
 
     def list_stream_states(self, stream_ids: list):
@@ -797,6 +874,7 @@ class LogicHubAPI:
         for s in stream_ids:
             new_stream_id_list.append(f'stream-{s}')
         body = {"streams": new_stream_id_list}
+        self.log.debug("Fetching stream states")
         response = self._http_request(method="POST", url=self.url.stream_states, body=body)
         return response.json()
 
@@ -826,10 +904,44 @@ class LogicHubAPI:
             "sortOrder": "ASC"
         }
 
+        self.log.debug("Fetching streams")
         response = self._http_request(method="POST", url=self.url.streams, body=body_dict, headers=headers, params=params)
         return response.json()
 
+    def list_user_groups(self, limit=None, hide_inactive=False):
+        limit = limit if limit and isinstance(limit, int) else 99999
+        params = {"pageSize": limit, "after": 0}
+        body = {"filters": []}
+        if hide_inactive:
+            body['filters'].append({"hideDeleted": True})
+        self.log.debug("Fetching user groups")
+        response = self._http_request(
+            url=self.url.user_groups,
+            method="POST",
+            headers={"Content-Type": "application/json"},
+            params=params,
+            body=body
+        )
+        return response.json()
+
+    def list_users(self, limit=None, hide_inactive=False):
+        limit = limit if limit and isinstance(limit, int) else 99999
+        params = {"pageSize": limit, "after": 0}
+        body = {"filters": []}
+        if hide_inactive:
+            body['filters'].append({"hideInactive": True})
+        self.log.debug("Fetching users")
+        response = self._http_request(
+            url=self.url.users,
+            method="POST",
+            headers={"Content-Type": "application/json"},
+            params=params,
+            body=body
+        )
+        return response.json()
+
     def list_workflows(self):
+        self.log.debug("Fetching workflows")
         response = self._http_request(method="GET", url=self.url.case_status_list_workflows)
         return response.json()
 
@@ -844,18 +956,48 @@ class LogicHubAPI:
             batch_id = int(re.sub(r'\D+', '', str(batch_id)))
         headers = {"Content-Type": "application/json"}
         body = {"method": "reschedule", "parameters": {}}
+        self.log.debug("Issuing command to reprocess batch")
         response = self._http_request(method="POST", url=self.url.batch_reprocess.format(batch_id), headers=headers, body=body)
         return response.json()
 
     def me(self):
+        self.log.debug("Fetching current user info")
         response = self._http_request(method="GET", url=self.url.me, timeout=30)
+        result = response.json()
+        if not all((
+            result.get('result'),
+            result['result']['id'],
+            result['result']['username'],
+            result['result']['role'],
+            result['result']['preferences']
+        )):
+            raise exceptions.validation.ResponseValidationError(input_var=response, message="API response does not match the expected schema for user profile and preferences")
+
+        self.__user_role = result['result']['role']
+        if result['result']:
+            new_id = result['result']['id']
+            if isinstance(new_id, str):
+                new_id = int(re.sub(r'\D+', '', new_id))
+            if new_id != self.__user_id:
+                self.__user_id = new_id
+            # Username will only exist already if password auth is used. In case of API token auth, capture and record the username.
+            if self.__username != result['result']['username']:
+                self.__username = result['result']['username']
+                self.log.debug(f"Current user updated to {self.__username} [ID: {self.__user_id}, role: {self.__user_role}]")
+        return result
+
+    def update_current_user_preferences(self, preferences):
+        self.log.debug("Patching user (updating preferences)")
+        response = self._http_request(method="PATCH", url=self.url.user.format(self.session_user_id), body={"preferences": preferences})
         return response.json()
 
     def case_prefix_refresh(self):
         _ = self.cases_get_prefix()
         return self.case_prefix
 
-    # ToDo Revisit eventually. Aborted on 2022-02-22 because it turns out this can be set w/ the case management integration as a custom field
+    # ToDo Revisit eventually. Aborted on 2022-02-22 because it turns out this
+    #  can be set w/ the case management integration as a custom field, but
+    #  there is still value in finishing this.
     # def case_link_alerts(self, case_id, alert_ids):
     #     if not isinstance(alert_ids, list):
     #         alert_ids = [alert_ids]
@@ -870,6 +1012,7 @@ class LogicHubAPI:
     #
     #     field_id = self.system_field_lh_linked_alerts.get('id')
     #     body = {"fields": [{"id": field_id, "value": alert_ids}]}
+    #     self.log.debug("Linking alert to case")
     #     response = self._http_request(
     #         method="PATCH",
     #         url=self.url.case_update_linked_alerts.format(case_id=case_id),
@@ -878,6 +1021,7 @@ class LogicHubAPI:
     #     output = response.json()
 
     def cases_get_prefix(self):
+        self.log.debug("Fetching case prefix")
         response = self._http_request(
             url=self.url.cases_get_prefix,
             headers={"Content-Type": "application/json"}
@@ -887,11 +1031,11 @@ class LogicHubAPI:
         return output
 
     def case_list_attached_notebooks(self, case_id, results_only=False):
-        case_id = helpers.format_case_id_with_prefix(case_id, self.case_prefix)
         body = {
-            "id": case_id,
+            "id": helpers.format_case_id_with_prefix(case_id, self.case_prefix),
             "key": "case",
         }
+        self.log.debug("Fetching attached notebooks")
         response = self._http_request(
             url=self.url.notebooks_attached,
             method="POST",
@@ -909,14 +1053,14 @@ class LogicHubAPI:
 
     def case_overwrite_attached_notebooks(self, case_id, notebooks):
         notebooks = helpers.format_notebook_ids(notebooks)
-        case_id = helpers.format_case_id_with_prefix(case_id, self.case_prefix)
         body = {
             "notebookAttachmentIds": notebooks,
             "notebookAttachedEntityId": {
                 "key": "case",
-                "id": case_id,
+                "id": helpers.format_case_id_with_prefix(case_id, self.case_prefix),
             }
         }
+        self.log.debug("Updating attached notebooks")
         response = self._http_request(
             url=self.url.notebooks_attach,
             method="PATCH",
@@ -926,40 +1070,11 @@ class LogicHubAPI:
         )
         return response.json()
 
-    def user_groups(self, limit=None, hide_inactive=False):
-        limit = limit if limit and isinstance(limit, int) else 99999
-        params = {"pageSize": limit, "after": 0}
-        body = {"filters": []}
-        if hide_inactive:
-            body['filters'].append({"hideDeleted": True})
-        response = self._http_request(
-            url=self.url.user_groups,
-            method="POST",
-            headers={"Content-Type": "application/json"},
-            params=params,
-            body=body
-        )
-        return response.json()
-
-    def users(self, limit=None, hide_inactive=False):
-        limit = limit if limit and isinstance(limit, int) else 99999
-        params = {"pageSize": limit, "after": 0}
-        body = {"filters": []}
-        if hide_inactive:
-            body['filters'].append({"hideInactive": True})
-        response = self._http_request(
-            url=self.url.users,
-            method="POST",
-            headers={"Content-Type": "application/json"},
-            params=params,
-            body=body
-        )
-        return response.json()
-
     def alert_fetch(self, alert_id, return_raw=False, **kwargs):
         alert_id = helpers.format_alert_id(alert_id)
         url = self.url.alert_fetch.format(alert_id)
         # Send request, but disable immediate testing in order to do specialized testing first
+        self.log.debug("Fetching alert")
         response = self._http_request(method="GET", url=url, test_response=False, **kwargs)
         if return_raw:
             return response
@@ -972,6 +1087,7 @@ class LogicHubAPI:
         if query:
             # First confirm that the query is valid
             body = {"queryType": "advanced", "query": query}
+            self.log.debug("Validating search syntax")
             response = self._http_request(
                 url=self.url.alerts_search_validate,
                 method="POST",
@@ -981,9 +1097,9 @@ class LogicHubAPI:
             response = response.json()
             is_invalid = response['result'].get('error')
             if is_invalid is None:
-                raise exceptions.Validation.BaseValidationError('Unexpected response while validating query string.')
+                raise exceptions.app.UnexpectedOutput('Unexpected response while validating query string.')
             elif is_invalid:
-                raise exceptions.Validation.AlertQueryValidationError(response['result'].get('message'))
+                raise exceptions.validation.AlertQueryValidationError(response['result'].get('message'))
 
     def alerts_search_advanced(self, query: str = None, limit: int = None):
         # Sanitize inputs
@@ -999,6 +1115,7 @@ class LogicHubAPI:
             "sortOrder": "desc",
             "query": query
         }
+        self.log.debug("Searching alerts")
         response = self._http_request(
             url=self.url.alerts_search_advanced,
             method="POST",
@@ -1010,11 +1127,13 @@ class LogicHubAPI:
     def stream_pause(self, stream_id, **kwargs):
         # Sanitize stream ID
         body = [f"stream-{helpers.format_stream_id(stream_id)}"]
+        self.log.debug("Pausing stream")
         response = self._http_request(method="POST", url=self.url.stream_pause, body=body, **kwargs)
         return response.json()
 
     def stream_resume(self, stream_id, **kwargs):
         # Sanitize stream ID
         body = [f"stream-{helpers.format_stream_id(stream_id)}"]
+        self.log.debug("Resuming stream")
         response = self._http_request(method="POST", url=self.url.stream_resume, body=body, **kwargs)
         return response.json()
