@@ -2,9 +2,13 @@ import json
 import re
 from copy import deepcopy
 
-from lhub.api import LogicHubAPI
-from lhub.log import Logger
-from lhub.common.helpers import format_notebook_ids
+from .api import LogicHubAPI
+from .log import Logger
+from .common.helpers import format_notebook_ids
+from .common.time import epoch_time_to_str
+from .exceptions.app import BaseAppError
+from .exceptions.validation import InputValidationError, ResponseValidationError
+from .exceptions.formatting import InvalidWorkflowIdFormat
 
 
 class Actions:
@@ -16,46 +20,6 @@ class Actions:
         # If this class has not been given a logger by the time it is instantiated, inherit the same one from LogicHubAPI
         self.log = self.log or self.__api.log
 
-    @staticmethod
-    def _reformat_alert_simple(alert: dict):
-        if not alert:
-            return {}
-        alert = deepcopy(alert)
-        # From this: {"displayName": "batch_start_millis", "value": "1632146400000"}
-        # To this: {"batch_start_millis": "1632146400000"}
-        # ToDo 2022-03-09, research further and verify since m92: the step below causes a failure now, and it looks like it's because the alert API is returning cleaner results now
-        # alert['additionalFields'] = {x['displayName']: x['value'] for x in alert.get('additionalFields', {}) if x.get('displayName')}
-
-        # From this: {"caseFieldId": "field-17", "displayName": "Alert Context", "fieldType": "Text", "value": "mdr_test"}
-        # To this: {"Alert Context": "mdr_test"}
-        for k in alert.get('mappedAlertFieldValues', {}):
-            if k.get('fieldType') == 'JSON':
-                try:
-                    k['value'] = json.loads(k['value'])
-                except (ValueError, TypeError, json.decoder.JSONDecodeError):
-                    pass
-        alert['mappedAlertFieldValues'] = {x['displayName']: x['value'] for x in alert.get('mappedAlertFieldValues', {}) if x.get('displayName')}
-        return alert
-
-    def _reformat_cmd_results(self, response_dict, drop_hidden_columns=True):
-        full_result = response_dict.copy()
-
-        result_raw = full_result["result"]
-        warnings = full_result["result"].get("warnings")
-        result_with_schema = result_raw["rows"]["data"]
-
-        # execution_ids = [x["id"] for x in result_with_schema]
-
-        if drop_hidden_columns:
-            result_output = [{k: v for k, v in _result['fields'].items() if k not in ["lhub_id", "lhub_page_num"]} for _result in result_with_schema]
-        else:
-            result_output = [{k: v for k, v in _result['fields'].items()} for _result in result_with_schema]
-
-        for _warning in warnings:
-            self.log.debug(f"WARNING RETURNED: {_warning}")
-
-        return result_output, warnings
-
     @staticmethod
     def __enrich_alert_data(alert: dict, included_standard_fields=None, included_additional_fields=None):
         # Verify and sanitize inputs
@@ -100,22 +64,59 @@ class Actions:
                 if raise_errors is True:
                     action_description = f" [{action_description}]" if action_description else ''
                     err_msg = f"Response{action_description} did not match the expected JSON schema. Missing expected key: {field}"
-                    raise ValueError(err_msg)
+                    raise ResponseValidationError(input_var=err_msg, message=action_description)
                 return False
             _dict = _dict[field]
         return True
 
+    @property
+    def playbook_ids(self):
+        response = self.list_playbooks()
+        return {p["id"]["id"]: p["name"] for p in response}
+
+    def __reformat_cmd_results(self, response_dict, drop_hidden_columns=True):
+        full_result = response_dict.copy()
+        result_raw = full_result["result"]
+        warnings = full_result["result"].get("warnings")
+        result_with_schema = result_raw["rows"]["data"]
+
+        if drop_hidden_columns:
+            result_output = [{k: v for k, v in _result['fields'].items() if k not in ["lhub_id", "lhub_page_num"]} for _result in result_with_schema]
+        else:
+            result_output = [{k: v for k, v in _result['fields'].items()} for _result in result_with_schema]
+
+        for _warning in warnings:
+            self.log.debug(f"WARNING RETURNED: {_warning}")
+        return result_output, warnings
+
     def execute_command(self, command_name, input_dict, reformat=True, result_limit=None):
         response = self.__api.execute_command({"command": command_name, "parameterValues": input_dict, "limit": result_limit})
         if not reformat:
             return response
 
-        response, _ = self._reformat_cmd_results(response)
+        response, _ = self.__reformat_cmd_results(response)
         return response
 
+    @staticmethod
+    def __reformat_alert_simple(alert: dict):
+        if not alert:
+            return {}
+        alert = deepcopy(alert)
+
+        # From this: {"caseFieldId": "field-17", "displayName": "Alert Context", "fieldType": "Text", "value": "mdr_test"}
+        # To this: {"Alert Context": "mdr_test"}
+        for k in alert.get('mappedAlertFieldValues', {}):
+            if k.get('fieldType') == 'JSON':
+                try:
+                    k['value'] = json.loads(k['value'])
+                except (ValueError, TypeError, json.decoder.JSONDecodeError):
+                    pass
+        alert['mappedAlertFieldValues'] = {x['displayName']: x['value'] for x in alert.get('mappedAlertFieldValues', {}) if x.get('displayName')}
+        return alert
+
     def get_alert_by_id(self, alert_id, simple_format=False, included_standard_fields=None, included_additional_fields=None):
         result = self.__api.alert_fetch(alert_id)
-        _ = self._result_dict_has_schema(result, "result", action_description="list users", raise_errors=True)
+        _ = self._result_dict_has_schema(result, "result", action_description="fetch alert", raise_errors=True)
 
         # Reformat alert and enrich as needed
         output = self.__enrich_alert_data(
@@ -126,14 +127,12 @@ class Actions:
 
         # Simplify mapped and additional field output
         if simple_format:
-            output = self._reformat_alert_simple(output)
-            # field_map = self.action_list_fields(map_mode="name")
-            # output['additionalFields'] = {}
+            output = self.__reformat_alert_simple(output)
         return output
 
     def get_batch_results_by_id(self, batch_id: int, limit=1000, keep_additional_info=False):
         result = self.__api.get_batch_results_by_id(batch_id=batch_id, limit=limit)
-        _ = self._result_dict_has_schema(result, "result", "data", raise_errors=True, action_description="fetch batch results")
+        _ = self._result_dict_has_schema(result, "result", "data", raise_errors=True, action_description="get batch results by batch ID")
         result = result["result"]["data"]
         if not keep_additional_info:
             result = [r["columns"] for r in result]
@@ -168,7 +167,6 @@ class Actions:
         _ = self._result_dict_has_schema(response, "result", action_description="get notebooks from case", raise_errors=True)
         return response.get('result', [])
 
-    # ToDo This finally works as of m91. Add an action for it.
     def get_stream_by_id(self, stream_id: int):
         result = self.__api.get_stream_by_id(stream_id)
         return result["result"]
@@ -180,11 +178,11 @@ class Actions:
         if isinstance(workflow_id, str):
             workflow_id = re.findall(r"(\d+)", workflow_id)
             if not workflow_id:
-                raise ValueError("Invalid Workflow ID: no numeric value found")
+                raise InvalidWorkflowIdFormat(input_var=workflow_id)
             workflow_id = workflow_id[0]
         workflow_id = int(workflow_id)
         result = self.__api.get_workflow_by_id(workflow_id=workflow_id)
-        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="fetch workflow by ID")
+        _ = self._result_dict_has_schema(result, "result", raise_errors=True, action_description="get workflow by ID")
         return result["result"]
 
     def list_baselines(self):
@@ -194,7 +192,7 @@ class Actions:
         baselines = result['data']
         stream_ids = [int(b['id']['id']) for b in baselines]
         stream_state_response = self.__api.list_stream_states(stream_ids=stream_ids)
-        _ = self._result_dict_has_schema(stream_state_response, "result", "streams", raise_errors=True, action_description="fetch stream states")
+        _ = self._result_dict_has_schema(stream_state_response, "result", "streams", raise_errors=True, action_description="list stream states")
         state_map = {_stream['streamId']: _stream['status'] for _stream in stream_state_response['result']['streams']}
         for n in range(len(baselines)):
             b = baselines[n]
@@ -202,6 +200,27 @@ class Actions:
             baselines[n]['baseline_config_status'] = state_map.get(f"stream-{b['id']['id']}")
         return result
 
+    @staticmethod
+    def __reformat_command_simple(result):
+        result = {
+            "name": result["name"],
+            "id": int(result["id"]["id"]),
+            "flowId": result["flowId"],
+            "owner": result["owner"],
+            "last_updated": epoch_time_to_str(result["lastUpdated"] / 1000),
+            "command_status": result["commandStatus"],
+        }
+        return result
+
+    def list_commands(self, filters: list = None, limit: int = None, simple_format=False):
+        result = self.__api.list_commands(filters=filters, limit=limit, offset=0)
+        _ = self._result_dict_has_schema(result, "result", "data", "data", action_description="list commands", raise_errors=True)
+        results = result["result"]["data"]
+        self.log.debug(f"{len(results)} commands found")
+        if simple_format:
+            results = [self.__reformat_command_simple(r) for r in results['data']]
+        return results
+
     def list_connections(self, filters=None, add_status=False):
         result = self.__api.list_connections(filters=filters)
         _ = self._result_dict_has_schema(result, "result", "data", action_description="list connections", raise_errors=True)
@@ -218,7 +237,7 @@ class Actions:
             connection_id = str(connection["id"]["id"])
             connection["status"] = statuses.get(connection_id)
             if not connection["status"]:
-                raise ValueError(f"Connection ID {connection_id} missing from status list")
+                raise BaseAppError(f"Connection ID {connection_id} missing from status list")
         return result
 
     def list_custom_lists(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = None):
@@ -247,9 +266,9 @@ class Actions:
 
     def list_fields(self, map_mode=None):
         assert not map_mode or map_mode in ['id', 'name'], f'Invalid output format: {map_mode}'
-        result = self.__api.fields
-        _ = self._result_dict_has_schema(result, "result", "data", raise_errors=True, action_description="list fields")
-        output = result["result"]["data"]
+        response = self.__api.fields
+        _ = self._result_dict_has_schema(response, "result", "data", raise_errors=True, action_description="list fields")
+        output = response["result"]["data"]
         if map_mode == "id":
             output = {f['id']: {k: v for k, v in f.items() if k != 'id'} for f in output}
         elif map_mode == "name":
@@ -259,14 +278,12 @@ class Actions:
     def list_integrations(self, name_filter: str = None, filter_type="contains"):
         filter_type = filter_type.lower()
         assert filter_type in ["equals", "contains"], f"Invalid filter type \"{filter_type}\""
-        assert isinstance(name_filter, str), "name_filter must be a string and cannot be None"
 
-        _response = self.__api.list_integrations()
-
-        # _query = _response.pop("query")
-        _result = _response.pop("result")
-        categories = _result["categories"]
-        results = _result["objects"]
+        response = self.__api.list_integrations()
+        _ = self._result_dict_has_schema(response, "result", "objects")
+        result = response.pop("result")
+        categories = result["categories"]
+        results = result["objects"]
 
         if name_filter:
             if filter_type == "contains":
@@ -297,15 +314,16 @@ class Actions:
         _ = self._result_dict_has_schema(result, "result", "data", "data", action_description="list notebooks", raise_errors=True)
         return result["result"]["data"]
 
-    @property
-    def playbook_ids(self):
-        response = self.list_playbooks()
-        return {p["id"]["id"]: p["name"] for p in response}
-
-    def list_playbooks(self, limit=None):
+    def list_playbooks(self, limit=None, map_mode=None):
+        assert not map_mode or map_mode in ['id', 'name'], f'Invalid output format: {map_mode}'
         response = self.__api.list_playbooks(limit=limit)
         _ = self._result_dict_has_schema(response, "result", "data", "data")
-        return response["result"]["data"]["data"]
+        output = response["result"]["data"]["data"]
+        if map_mode == "id":
+            output = {f['id']['id']: {k: v for k, v in f.items() if k != 'id'} for f in output}
+        elif map_mode == "name":
+            output = {f['name']: {k: v for k, v in f.items() if k != 'name'} for f in output}
+        return output
 
     def list_streams(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = 0):
         result = self.__api.list_streams(search_text=search_text, filters=filters, limit=limit, offset=offset)
@@ -313,14 +331,34 @@ class Actions:
         return result["result"]["data"]["data"]
 
     def list_user_groups(self, hide_inactive=False):
-        result = self.__api.user_groups(hide_inactive=hide_inactive)
+        result = self.__api.list_user_groups(hide_inactive=hide_inactive)
         _ = self._result_dict_has_schema(result, "result", "data", action_description="list user groups", raise_errors=True)
         return result["result"]
 
-    def list_users(self, hide_inactive=False):
-        result = self.__api.users(hide_inactive=hide_inactive)
+    @staticmethod
+    def __reformat_user_simple(user: dict):
+        groups = [g['name'] for g in user['groups'] if not g.get("isDeleted", False)]
+        user_attributes = {
+            "username": user.get("name"),
+            "is_admin": user["role"]["value"] == "admin",
+            "groups": ', '.join(groups),
+            "email": user.get("email"),
+            "is_deleted": user.get("isDeleted"),
+            "is_enabled": user.get("isEnabled"),
+            "auth_type": user.get("authenticationType"),
+            "id": user.get("userId"),
+        }
+        # would it be beneficial to make a class object for these instead of returning a dict?
+        return user_attributes
+
+    def list_users(self, hide_inactive=True, simple_format=False):
+        result = self.__api.list_users(hide_inactive=hide_inactive)
         _ = self._result_dict_has_schema(result, "result", "data", action_description="list users", raise_errors=True)
-        return result["result"]
+        results = result["result"]
+        self.log.debug(f"{len(results['data'])} users found")
+        if simple_format:
+            results = [self.__reformat_user_simple(result) for result in results['data']]
+        return results
 
     def list_workflows(self):
         result = self.__api.list_workflows()
@@ -364,7 +402,7 @@ class Actions:
             if fetch_details == 'simple':
                 simple_format = True
             elif fetch_details != 'standard':
-                raise ValueError(f'Invalid input for fetch_details: {fetch_details}')
+                raise InputValidationError(input_var=fetch_details, action_description="fetch_details")
         query = query.strip() if query and query.strip() else ""
         result = self.__api.alerts_search_advanced(query=query, limit=limit)
         _ = self._result_dict_has_schema(result, "result", action_description="search alerts", raise_errors=True)
@@ -387,11 +425,36 @@ class Actions:
         _ = self._result_dict_has_schema(result, "result", action_description="resume stream", raise_errors=True, accept_null=True)
         return result
 
+    def update_current_user_preferences(self, **kwargs):
+        kwargs = {k: v for k, v in kwargs.items() if v is not None}
+        if not kwargs:
+            raise InputValidationError(
+                input_var=kwargs, action_description="updating user preferences",
+                message=f"Invalid input for updating user preferences. At least one ID with boolean value is required, but none were received."
+            )
+        for k, v in kwargs.items():
+            if not isinstance(v, bool):
+                raise InputValidationError(input_var=v, action_description="updating user preference", message=f"Invalid input for updating user preferences. Preference values must be boolean but received: {v}")
+        preferences = self.__api.me()['result']['preferences']
+        preference_ids = [p['id'] for p in preferences]
+        for k in kwargs:
+            if k not in preference_ids:
+                raise InputValidationError(input_var=k, action_description="updating user preference", message=f"Invalid preference ID: {k}. Valid IDs are: {', '.join(preference_ids)}")
+        for n in range(len(preferences)):
+            if preferences[n]['id'] in kwargs:
+                preferences[n]['value'] = kwargs[preferences[n]['id']]
+
+        return self.__api.update_current_user_preferences(preferences)
+
 
 class LogicHub:
     verify_ssl = True
 
-    def __init__(self, hostname, api_key=None, username=None, password=None, verify_ssl=True, cache_seconds=None, verify_api_auth=True, logger: Logger = None, log_level=None, default_timeout=None, **kwargs):
+    def __init__(
+            self, hostname, api_key=None, username=None, password=None,
+            verify_ssl=True, cache_seconds=None, verify_api_auth=True,
+            logger: Logger = None, log_level=None, default_timeout=None,
+            **kwargs):
         # If the LogicHubAPI class object has not been given a logger by the time this class is instantiated, set one for it
         LogicHubAPI.log = LogicHubAPI.log or logger or Logger()
         if log_level:
